
# SPDX-License-Identifier: BSD-2
# this file is autogenerated during the build

_versions = {
    "tss2-esys": "4.1.3",
    "tss2-fapi": "4.1.3",
    "tss2-policy": "4.1.3",
    "tss2-tcti-spi-helper": "4.1.3",

}
